/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File mouse.c : Emulation of the SAM mouse
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <stdlib.h>

#include "z80.h"
#include "mouse.h"
#include "sam.h"


/* Buffer used to store the Mouse data */

typedef union _MouseBuff {
	struct _MouseStruct {
		uns8 MouseStrobe;
		uns8 MouseDummy;
		uns8 Button_status;
		uns8 Y256;
		uns8 Y16;
		uns8 Y1;
		uns8 X256;
		uns8 X16;
		uns8 X1;
	}    MouseStruct;
	uns8 MouseArray[9]; 
} MouseBuff;

int MouseStrobed = 0;

static MouseBuff *TheMouseBuff;

static uns8 *BuffPtr;
static int  ByteCount = 0;	


void 
sam_mouse_init()
{
	TheMouseBuff = (MouseBuff *)malloc(sizeof(MouseBuff));
	sam_mouse_reset(1);	
}

uns8 
sam_mouse_in()
{
	uns8 in_byte;

	if (!ByteCount)
		MouseStrobed = 1;

	if (ByteCount++ < 9) {
		in_byte = *BuffPtr++;
	}
	else {
		in_byte = 0xff;		/* dummy */
		sam_mouse_reset(0); 
		sam_mouse_xy_write(0,0);
		ByteCount++;		/* Skip the strobe */
		BuffPtr++;
	}
/*	printf("sam_mouse_in()=%d\n",in_byte);	*/
	return(in_byte); 
}

void 
sam_mouse_reset(int ResetButtons) 
{
/* Clear the counters for the mouse */

	int i;

	ByteCount = 0;
	BuffPtr = TheMouseBuff->MouseArray;

	*BuffPtr = *(BuffPtr + 1) = 0xff;
	if (ResetButtons) 
		*(BuffPtr + 2) = 0xff;

	/* There is never a need to actually clear the position
	   counters - as this is done when the pointer is warped
	   back to the center of the window - ie we always get
	   a sam_mouse_xy_write(0,0) event                         */

}

void 
sam_mouse_xy_write(int DeltaX, int DeltaY)
{
/* Write the X,Y change DeltaX, DeltaY into the mouse buffer */

/*	printf("sam_mouse_xy_write(%d,%d)\n",DeltaX,DeltaY); */
	TheMouseBuff->MouseStruct.Y256 = (DeltaY & 0x0f00) >> 8;
	TheMouseBuff->MouseStruct.Y16  = (DeltaY & 0x00f0) >> 4;
	TheMouseBuff->MouseStruct.Y1   = (DeltaY & 0x000f);
	TheMouseBuff->MouseStruct.X256 = (DeltaX & 0x0f00) >> 8;
	TheMouseBuff->MouseStruct.X16  = (DeltaX & 0x00f0) >> 4;
	TheMouseBuff->MouseStruct.X1   = (DeltaX & 0x000f);
}

void 
sam_mouse_button_write(int ButtonData)
{
/* Write the Button status into mouse buffer */

	TheMouseBuff->MouseStruct.Button_status = (uns8)ButtonData;	
}

