/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File processor.c : z80 main loop and Sam interrupt routines
 *
 * Intel 80x86 assembly optimisations Copyright (c) 1996 Allan Skillman
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <signal.h>
#include <sys/types.h>

#include "z80.h"
#include "mem.h"
#include "resource.h"
#include "util.h"
#include "debug.h"
#include "sam.h"
#include "mouse.h"

#ifndef __GNUC__
#define inline
#endif

#define parity(a) (partable[a])

uns8 partable[256]={
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
      4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4
   };


int disass;

int instrs;			/* no. of instructions executed since powerup */
int n_interrupts;		/* no. of interrupts since powerup */

static int LineFreq;		/* No of instructions / line generated */
static int IntActiveTime;	/* Time (in instructions) an int is active (Port 249 <> 0xff) */
int LineNo; 			/* Scan line being generated */
int InterruptType;		/* Type of z80 interupt to deal with */


static int MouseActiveTime;	/* Time (in instructions) from a mouse strobe to a counter reset */

static int DelayedEI = 0;	/* Flag and counter to carry out a delayed */

int flash_phase = 0;			/* gets toggled between 1 and 0 twice per second */

static void nop(void);

#ifdef SLOWDOWN
/* Do nothing! */
static void
nop(void)
{
}
#endif


/*
 * The main z80 emulation loop
 */
void
emulator_main_loop(void)
{

   register uns8 a, b, c, f, d, e, h, l;

   uns8 r, a1, f1, b1, c1, d1, e1, h1, l1, i, iff1, iff2, im;
   register uns16 pc;
   uns16 ix, iy, sp;
   register unsigned int radjust;
   register uns8 ixoriy, new_ixoriy;
   register uns8 op;

/* Reset the CPU */

	int refresh_interval;

#ifdef SLOWDOWN
	int islow;
#endif

	register int IntActiveCounter;		/* Counter used for int Active timing */
	register int MouseActiveCounter;	/* Counter used for mouse Active timing */

	/* from (vga/x)samgrx.c */

	extern void (*line_update)(int);
	extern void (*borderline_update)(void);

	extern int  BorderTop;
	extern int  BorderBottom;
	extern int  ClutChangeInFrame;
	extern int  BorderChangeInFrame;
	extern int  VmprChangeInFrame;	

	/* from vga_samio.c */
	extern void check_events(void);

	/* from mouse.c */
	
	extern int MouseStrobed;

	a=f=b=c=d=e=h=l=a1=f1=b1=c1=d1=e1=h1=l1=i=r=iff1=iff2=im=0;
	ixoriy=new_ixoriy=0;
	ix=iy=sp=pc=0;
	radjust=0;

	InterruptType = 0;
	
	disass = 0;

	refresh_interval = GetCfg(rr);

	/* start interrupts */

	/* Calculate the mean no. of instructions/line generated */

	LineFreq = GetCfg(ifreq) / (60. + 192.);

	/* Calculate the mean no. of instructions executed in the 100usecs
	   an interrupt remains active */
	   
	IntActiveTime = GetCfg(ifreq) * (100.e-6 / 0.02);
	simcoupe_mesg(SIMCOUPE_INFO,"LineFreq     : %d", LineFreq);
        simcoupe_mesg(SIMCOUPE_INFO,"IntActiveTime : %d", IntActiveTime);
	LineNo   = 0;	
	IntActiveCounter = IntActiveTime;

	/* Similar for mouse reset */	
	MouseActiveTime = IntActiveTime;
	MouseActiveCounter = MouseActiveTime;


	while(1) {
		instrs++;

		if (disass) {

			disass = run_debug(pc);
			if (disass == 2) {
				fprintf(stderr, "Z80 CPU State\n-------------\n");
				fprintf(stderr, "  AF  = %04X\t\t  AF' = %04X\t\t  A = %02X  F = %02X\n",
					af, af_prime, a, f);
				fprintf(stderr, "  BC  = %04X\t\t  BC' = %04X\t\t  B = %02X  C = %02X\n",
					bc, bc_prime, b, c);
				fprintf(stderr, "  DE  = %04X\t\t  DE' = %04X\t\t  D = %02X  E = %02X\n",
					de, de_prime, d, e);
				fprintf(stderr, "  HL  = %04X\t\t  HL' = %04X\t\t  H = %02X  L = %02X\n",
					hl, hl_prime, h, l);
				fprintf(stderr, "  IX  = %04X\t\t  IY  = %04X\n\n",
					ix, iy);
		
				fprintf(stderr, "  PC  = %04X\t\t  SP  = %04X\n",
					pc, sp);
	
				fprintf(stderr, "   R  = %02X\t\t   I  = %02X\n",
					r, i);
	
				fprintf(stderr, "  interrupts %s, IM%d\n",
					iff1 ? "enabled" : "disabled", im);
	
				fprintf(stderr, "  FLAGS: C=%d N=%d P/V=%d H=%d Z=%d S=%d\n\n",
					f & F_CARRY    ? 1 : 0,
					f & F_NADD     ? 1 : 0,
					f & F_OVERFLOW ? 1 : 0,
					f & F_HCARRY   ? 1 : 0,
					f & F_ZERO     ? 1 : 0,
					f & F_NEG      ? 1 : 0);
			}
		}

#ifdef SLOWDOWN
		for (islow = 0; islow < GetCfg(slowdown); islow++) nop();
#endif

		ixoriy=new_ixoriy;
		new_ixoriy=0;

		DelayedEI = 0;
		op=read_byte(pc);
		pc++;
		radjust++;
		switch(op){

#include "z80ops.c"

		}

		/* If we have strobed the mouse, we will need to do
		   a reset after 100 usecs */
		   
		if ((MouseStrobed) && (!MouseActiveCounter--)) {
			MouseActiveCounter = MouseActiveTime;
			MouseStrobed = 0;
			sam_mouse_reset(0);
		}



		/* On average there are about 10000 Z80 instructions in 0.02 secs, on
	  	 * a 6.0MHz Z80b (assuming average instruction length of 12 T-states)
		 */


		/* Call the interrupt if required */

		if (!DelayedEI && (status_reg != 0xff) && iff1) {			
/*			do_interrupt(op);  */
		         if(iff1){
 		           if(op==0x76)pc++;
      			   iff1=iff2=0;
    			   switch(im){
    			       	case 0: /* IM 0 */
 		              	case 1: /* undocumented */
            			case 2: /* IM 1 */
               			/* there is little to distinguish between these cases */
    				  push2(pc);
		                  pc=0x38;
                		  break;
		               case 3: /* IM 2 */
		                  {
	        	             uns16 addr=read_word((i<<8)|0xff);
		               	     push2(pc);
				     pc=addr;
	                  	  }
		           }
			 }
		}

		if ((status_reg != 0xff) && (!IntActiveCounter--)) {
			IntActiveCounter = IntActiveTime;
			status_reg = 0xff;
		}
		

		/* Deal with the display */
		if (!(instrs % LineFreq)) {
			
			if (LineNo == (191+60)) {
			/* FlyBack - Frame interrupt */

				LineNo = 0;
				IntActiveCounter = IntActiveTime;
				status_reg = 0xf7;
				
				n_interrupts++;
				
				check_events();
				FlyBack();

				/* every 25 frames -> twice/sec. */
				if (((n_interrupts % 25) == 0) && (VMPR_SCREEN_MODE < MODE_3) &&
				      GetCfg(flashing)) {

				/* Sam modes 1 and 2 use the flashing attribute bit. Every frame the
				   flash_phase is flipped (pen <-> paper in the screen write routines)
				   and ClutChangeInFrame is set to the very last display line to 
				   make sure the whole display is updated
				*/

					flash_phase = 1 - flash_phase;
					ClutChangeInFrame = (192+59);

				}

				/* Deal with NMI type interrupts */
		        	switch(InterruptType) {
		            	 case Z80_none:
		            		break;
		            	 case Z80_nmi:
					if(read_byte(pc)==0x76)pc++;
					iff2=iff1;
       					iff1=0;
        		     	  /* The Z80 performs a machine fetch cycle for 5 Tstates
               	 		  but ignores the result.  It takes a further 10 Tstates
               	   		  to jump to the NMI routine at 0x66. */
				  	push2(pc);
					pc=0x66;
					InterruptType = 0;
					break;
	            		 case Z80_reset:
					a=f=b=c=d=e=h=l=a1=f1=b1=c1=d1=e1=
					h1=l1=i=r=iff1=iff2=im=0;
					ix=iy=sp=pc=0;
					radjust=0;
					init_pages();
                       			init_ports();
					InterruptType = 0;
					break;
		    		 default:
		    			break;
			 	}

			}
			else if ((line_int < 192) && (LineNo == line_int+30)) {
			/* Line Interrupt */

				IntActiveCounter = IntActiveTime;
				status_reg = 0xfe;
			}
			
			if (LineNo <= 29) {
			/* Top Border */
				hpen = 0;
				attr_reg = 255;

				if (LineNo > BorderTop)
					borderline_update();

			}
			else if (LineNo > 29 && LineNo < 222) {
			/* Main Screen */
			
				line_update(LineNo-30);
				
				/* Update the lpen port */
				hpen = LineNo - 30;			
				attr_reg = 0;
			}
			else {
			/* Bottom Border */
				hpen = 192;
				/* Set attr port to 0xff in the border for
				   Spectrum compatability */
				attr_reg = 255;

				if (LineNo < BorderBottom)
					borderline_update();


			}
		
			/* Now we are Generating the next line down so
			   update the counter */
			   
			LineNo++;
			
			/* Reset the display flags if required */
			if (LineNo == VmprChangeInFrame)
				VmprChangeInFrame = False;
			if (LineNo == ClutChangeInFrame)
				ClutChangeInFrame = False;
			if (LineNo == BorderChangeInFrame)
				BorderChangeInFrame = False;
		
                }
	}

}
