/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File samio.c : SAM io hardware emulation
 *
 * Intel 80x86 assembly optimisations Copyright (c) 1996 Allan Skillman
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>

#include "z80.h"
#include "mem.h"
#include "sam.h"
#include "resource.h"
#include "util.h"
#include "version.h"
#include "mouse.h"
#include "fdi.h"

uns8 keyports[10];	/* 8 rows of keys (+ 1 row for unscanned keys) */
uns8 inPorts[256], outPorts[256];

/* All to do with screen painting */

int ClutChangeInFrame;			/* signals a clut change has been made in 
					   the present frame */
int BorderChangeInFrame;		/* signals a border change has been made in
					   the present frame */
int VmprChangeInFrame;			/* signals a vmpr change has been made in
					   the present frame */

extern int LineNo;
extern int SAMPointerMode;
extern unsigned long pixels[];                /* 128  SAM colours */

extern void (*line_update)(int);
extern void (*line_update_mode_1)(int);
extern void (*line_update_mode_2)(int);
extern void (*line_update_mode_3)(int);
extern void (*line_update_mode_4)(int);

void 
init_ports(void) 
{
	int i;

	vmpr = 0x81;
	hmpr = 0x01;
	lmpr = 0x1f;
	midi = 0x00;	
	border = 0x00;

	for (i=0; i < 16; i++)
		clut[i] = (uns8)pixels[i | (i & 0x07)<<4];
		
	keyboard = 0x5f;
	status_reg = 0xff;
	
	line_int = 0xff;
        hpen = 0x00;
        lpen = 0x00;
        	
        clock_reg = 0x00;

	lepr = 0x00;
	hepr = 0x00;
	attr_reg = 0xff;

}

static inline void 
out_hepr(uns8 val)
{
	int ExternalPage;

        hepr = val;

/* If somebody tries to access non existent external memory, page in the 
   ROM in sections C or D. This means that read-write tests on the
   memory will fail - as required
*/
	
	if (HMPR_MCNTRL) {
     
		if ((ExternalPage = NMAINPAGES+val) >= (ROM0-(GetCfg(externalMem)<<6)))
		   page_in(3, ExternalPage);
		else
	   	   page_in(3, ROM0);
	}
}

static inline void 
out_lepr(uns8 val)
{

	int ExternalPage;

        lepr = val;


/* If somebody tries to access non existent external memory, page in the 
   ROM in sections C or D. This means that read-write tests on the
   memory will fail - as required
*/
        
       	if (HMPR_MCNTRL) {

		if ((ExternalPage = NMAINPAGES+val) >= (ROM0-(GetCfg(externalMem)<<6)))
		   page_in(2, ExternalPage);
		else
		   page_in(2, ROM0);
	}
}

static inline void 
out_vmpr(uns8 val)
{
        vmpr = (val & 0x7f) | 0x80 ;

	switch (VMPR_SCREEN_MODE) {
	
	 case MODE_1:
	  line_update = line_update_mode_1;
	  break;
	 case MODE_2:
	  line_update = line_update_mode_2;
	  break;
	 case MODE_3:
	  line_update = line_update_mode_3;
	  break;
	 case MODE_4:
	  line_update = line_update_mode_4;
	  break;

	} /* End switch */
	if (!BORD_SOFF)
	 VmprChangeInFrame = LineNo;
}

static inline void 
out_hmpr(uns8 val)
{
        hmpr = val;
        
	if (LMPR_ROM1) 
	   page_in(3,ROM1);
	else
	   page_in(3,(HMPR_PAGE + 1)& 0x1f);
  
	page_in(2, HMPR_PAGE);

   
	if (HMPR_MCNTRL) {
		out_hepr(hepr);
		out_lepr(lepr);
	}

}


static inline void 
out_lmpr(uns8 val)
{
        lmpr = val;
        
	if (LMPR_RAM0)
	   page_in(0, LMPR_PAGE);
	else
	   page_in(0, ROM0);

        page_in(1, (LMPR_PAGE + 1) & 0x1f);

        if (LMPR_ROM1)
           page_in(3, ROM1);
        else
           page_in(3, (HMPR_PAGE + 1) &0x1f);
           
}

static inline void 
out_clut(uns16 port, uns8 val)
{
	uns8 old_val = clut[port & 0x0f];
  
	if (old_val != (uns8)pixels[val & 0x7f]) {
		clut[port & 0x0f] = (uns8)pixels[val & 0x7f];
	 	if (!BORD_SOFF)
			ClutChangeInFrame = LineNo;
	}
}

/* The Actual port input and output routines */

void
out_byte(uns16 port, uns8 val)
{
/*	printf("out_byte(%d,%d)\n",port,val); */
	/* border colour and (possibly) speaker control */
	switch (port & 0xff) {
	    case BORDER_PORT:
		if (BORD_VAL(border) != BORD_VAL(val)) {
			BorderChangeInFrame = LineNo;
		}
		if (BORD_SOFF && !(val  & 0x80)) 
		/* display turned back on - do we need to do 
		   something?? */;
		   		
		border = val;
		keyboard = (keyboard & 0x7f) |BORD_SOFF;
		break;
	    
	/* banked memory management */

            case VMPR_PORT:
		if ((vmpr & 0x7f) != (val & 0x7f)) out_vmpr(val);
        	break;
            case HMPR_PORT:
		if (hmpr != val) out_hmpr(val);
        	break;
            case LMPR_PORT:
        	if (lmpr != val) out_lmpr(val);
        	break;
	    case CLUT_BASE_PORT:
       		out_clut(port >> 8, val);
       		break;
        	
	/* External memory */

	    case HEPR_PORT:
        	out_hepr(val);
        	break;
	    case LEPR_PORT:
        	out_lepr(val);
        	break;
 	
	    case LINE_PORT:
		line_int = val;
		break;
	} /* End switch */

/* Disk IO
*/
        if (((port & 0xff) >= DISK1_PORT &&
             (port & 0xff) < DISK1_PORT + 8) ||
            ((port & 0xff) >= DISK2_PORT &&
             (port & 0xff) < DISK2_PORT + 8)) 
             	fdi_out(port,val);

#ifdef PSEUDO_IO
	if (port == PSEUDO_STDOUT)
		putc((int)val, stdout);
	else if (port == PSEUDO_STDERR)
		putc((int)val, stderr);
#endif

	outPorts[port & 0xff] = val;
}


uns8
in_byte(uns16 port)
{
	int res = 0x1f;
	int highbyte;

	if ((port & 0xff) < 0x10)
		return(0); 
        if ((port & 0x1ff) == HPEN_PORT)
                return(hpen); 

	switch (port & 0xff) {
 	    case KEYBOARD_PORT:  /* keyboard 1 / mouse */
		highbyte = port >> 8;
                if (highbyte ==  0xff)  {
                	res = keyports[8] & 0x1f;
			if (SAMPointerMode && (res == 0x1f)) {
				res = (int)(sam_mouse_in() & 0x1f);
			}
		}

		if (!(highbyte & 0x80)) res &= keyports[7] & 0x1f;
		if (!(highbyte & 0x40)) res &= keyports[6] & 0x1f;
		if (!(highbyte & 0x20)) res &= keyports[5] & 0x1f;
		if (!(highbyte & 0x10)) res &= keyports[4] & 0x1f;
		if (!(highbyte & 0x08)) res &= keyports[3] & 0x1f;
		if (!(highbyte & 0x04)) res &= keyports[2] & 0x1f;
		if (!(highbyte & 0x02)) res &= keyports[1] & 0x1f;
		if (!(highbyte & 0x01)) res &= keyports[0] & 0x1f;
		return (keyboard = (res | (keyboard & 0xe0)));
		break; 

	    case STATUS_PORT:  /* keyboard  2 */
		highbyte = port >> 8;
		res = 0xe0;
		if (!(highbyte & 0x80)) res &= keyports[7] & 0xe0;
		if (!(highbyte & 0x40)) res &= keyports[6] & 0xe0;
		if (!(highbyte & 0x20)) res &= keyports[5] & 0xe0;
		if (!(highbyte & 0x10)) res &= keyports[4] & 0xe0;
		if (!(highbyte & 0x08)) res &= keyports[3] & 0xe0;
		if (!(highbyte & 0x04)) res &= keyports[2] & 0xe0;
		if (!(highbyte & 0x02)) res &= keyports[1] & 0xe0;
		if (!(highbyte & 0x01)) res &= keyports[0] & 0xe0;
		return (((res & 0xe0) | (status_reg & 0x1f)));
		break;

	/* banked memory management */

            case VMPR_PORT:
        	return(vmpr);
        	break;
            case HMPR_PORT:
             	return(hmpr);
             	break;
            case LMPR_PORT:
        	return(lmpr);
		break;
            case CLOCK_PORT:
                return(clock_reg); 
		break;
            case LPEN_PORT:
                return(lpen); 
		break;
            case ATTR_PORT:
                return(attr_reg);
                break;
	/* External memory */
            case HEPR_PORT:
		return(hepr);
		break;
            case LEPR_PORT:
                return(lepr);
                break;
	} /* End switch */

/* Disk IO
*/
	
        if (((port & 0xff) >= DISK1_PORT &&
             (port & 0xff) < DISK1_PORT + 8) ||
            ((port & 0xff) >=  DISK2_PORT &&
             (port & 0xff) < DISK2_PORT + 8)) 
             	return(fdi_in(port));

        
#ifdef PSEUDO_IO
	if (port == PSEUDO_STDIN)
		return (uns8)getchar();
	else
#endif
			
	return 0xff;
}

/* Interface routine to do a out_vmpr call from outside this */
/* source file as the internal routine is static inline      */

void call_vmpr(uns8 vmpr_val)
{
	out_vmpr(vmpr_val);
}
